//
//  DirectSoundStreamer.cpp
//  BeebEmX
//
//  Created by Steven Bagley on 18/05/2015.
//  Copyright (c) 2015 Brokentooth. All rights reserved.
//

#include "DirectSoundStreamer.h"

DirectSoundStreamer::DirectSoundStreamer( std::size_t rate, bool primary ) : m_pDirectSound(), m_pDirectSoundBuffer(), m_begin(), m_physical( rate * 100 / 1000 ), m_rate( rate ), m_stream( rate * 20 / 1000 ), m_virtual( m_count * m_stream )
{
    // Create DirectSound
    if( FAILED( DirectSoundCreate( 0, &m_pDirectSound, 0 ) ) )
        throw Fail();
    HRESULT hResult( m_pDirectSound->SetCooperativeLevel( GETHWND, primary ? DSSCL_WRITEPRIMARY : DSSCL_PRIORITY ) );
    if( FAILED( hResult ) )
    {
        if( primary && hResult == DSERR_UNSUPPORTED )
            throw PrimaryUnsupported();
        throw Fail();
    }
    
    // Create DirectSoundBuffer
    WAVEFORMATEX wfx = {};
    wfx.wFormatTag = WAVE_FORMAT_PCM;
    wfx.nChannels = 1;
    wfx.nSamplesPerSec = DWORD( m_rate );
    wfx.wBitsPerSample = 8;
    wfx.nBlockAlign = wfx.nChannels * wfx.wBitsPerSample / 8;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
    
    DSBUFFERDESC dsbd = { sizeof( DSBUFFERDESC ) };
    dsbd.dwFlags = DSBCAPS_GETCURRENTPOSITION2;
    if( primary )
        dsbd.dwFlags |= DSBCAPS_PRIMARYBUFFER;
    else
    {
        if( m_physical < m_virtual )
            m_physical = m_virtual;
        dsbd.dwBufferBytes = DWORD( m_physical );
        dsbd.lpwfxFormat = &wfx;
    }
    dsbd.guid3DAlgorithm = DS3DALG_DEFAULT;
    if( FAILED( m_pDirectSound->CreateSoundBuffer( &dsbd, &m_pDirectSoundBuffer, 0 ) ) )
        throw Fail();
    
    if( primary )
    {
        // Setup primary buffer
        if( FAILED( m_pDirectSoundBuffer->SetFormat( &wfx ) ) )
            throw Fail();
        DSBCAPS dsbc = { sizeof( dsbc ) };
        if( FAILED( m_pDirectSoundBuffer->GetCaps( &dsbc ) ) )
            throw Fail();
        m_physical = std::size_t( dsbc.dwBufferBytes );
    }
    else
    {
        // Clear buffer
        void *p;
        DWORD size;
        if( FAILED( m_pDirectSoundBuffer->Lock( 0, 0, &p, &size, 0, 0, DSBLOCK_ENTIREBUFFER ) ) )
            throw Fail();
        using std::memset;
        memset( p, 128, std::size_t( size ) );
        if( FAILED( m_pDirectSoundBuffer->Unlock( p, size, 0, 0 ) ) )
            throw Fail();
    }
    
    // Start source voice
    Play();
}

void DirectSoundStreamer::Stream( void const *pSamples )
{
    // Verify buffer availability
    DWORD play, write;
    if( FAILED( m_pDirectSoundBuffer->GetCurrentPosition( &play, &write ) ) )
        return;
    if( write < play )
        write += DWORD( m_physical );
    std::size_t begin( m_begin );
    if( begin < play )
        begin += m_physical;
    if( begin < write )
        begin = std::size_t( write );
    std::size_t end( begin + m_stream );
    if( play + m_virtual < end )
        return;
    begin %= m_physical;
    
    // Copy samples to buffer
    void *ps[ 2 ];
    DWORD sizes[ 2 ];
    HRESULT hResult( m_pDirectSoundBuffer->Lock( DWORD( begin ), DWORD( m_stream ), &ps[ 0 ], &sizes[ 0 ], &ps[ 1 ], &sizes[ 1 ], 0 ) );
    if( FAILED( hResult ) )
    {
        if( hResult != DSERR_BUFFERLOST )
            return;
        m_pDirectSoundBuffer->Restore();
        if( FAILED( m_pDirectSoundBuffer->Lock( DWORD( begin ), DWORD( m_stream ), &ps[ 0 ], &sizes[ 0 ], &ps[ 1 ], &sizes[ 1 ], 0 ) ) )
            return;
    }
    using std::memcpy;
    memcpy( ps[ 0 ], pSamples, std::size_t( sizes[ 0 ] ) );
    if( ps[ 1 ] )
        memcpy( ps[ 1 ], static_cast< char const * >( pSamples ) + sizes[ 0 ], std::size_t( sizes[ 1 ] ) );
    m_pDirectSoundBuffer->Unlock( ps[ 0 ], sizes[ 0 ], ps[ 1 ], sizes[ 1 ] );
    
    // Select next buffer
    m_begin = end % m_physical;
}




XAudio2Streamer::XAudio2Streamer( std::size_t rate ) : m_samples(), m_pXAudio2(), m_pXAudio2MasteringVoice(), m_pXAudio2SourceVoice(), m_index(), m_rate( rate ), m_size( rate * 20 / 1000 )
{
    // Allocate memory for buffers
    m_samples.resize( m_count * m_size, 128 );
    
    // Create XAudio2
    if( FAILED( XAudio2Create( &m_pXAudio2, 0, XAUDIO2_DEFAULT_PROCESSOR ) ) )
        throw Fail();
    
    // Create mastering voice
    if( FAILED( m_pXAudio2->CreateMasteringVoice( &m_pXAudio2MasteringVoice ) ) )
        throw Fail();
    
    // Create source voice
    WAVEFORMATEX wfx = {};
    wfx.wFormatTag = WAVE_FORMAT_PCM;
    wfx.nChannels = 1;
    wfx.nSamplesPerSec = DWORD( m_rate );
    wfx.wBitsPerSample = 8;
    wfx.nBlockAlign = wfx.nChannels * wfx.wBitsPerSample / 8;
    wfx.nAvgBytesPerSec = wfx.nSamplesPerSec * wfx.nBlockAlign;
    if( FAILED( m_pXAudio2->CreateSourceVoice( &m_pXAudio2SourceVoice, &wfx, 0, XAUDIO2_DEFAULT_FREQ_RATIO ) ) )
        throw Fail();
    
    // Start source voice
    Play();
}

void XAudio2Streamer::Stream( void const *pSamples )
{
    // Verify buffer availability
    XAUDIO2_VOICE_STATE xa2vs;
    m_pXAudio2SourceVoice->GetState( &xa2vs );
    if( xa2vs.BuffersQueued == m_count )
        return;
    
    // Copy samples to buffer
    Sample *pBuffer( &m_samples[ m_index * m_size ] );
    using std::memcpy;
    memcpy( pBuffer, pSamples, m_size );
    
    // Submit buffer to voice
    XAUDIO2_BUFFER xa2b = {};
    xa2b.AudioBytes = UINT32( m_size );
    xa2b.pAudioData = pBuffer;
    if( FAILED( m_pXAudio2SourceVoice->SubmitSourceBuffer( &xa2b ) ) )
        return;
    
    // Select next buffer
    m_index = ( m_index + 1 ) % m_count;
}


